import logging
import time
import datetime
from nubium_utils.custom_exceptions import NoMessageError, ConsumeMessageError, IncompatibleEnvironmentError
from nubium_utils.metrics import MetricsManager
from .confluent_runtime_vars import env_vars


LOGGER = logging.getLogger(__name__)


def _wait_until_message_time(msg_timestamp, guid):
    """
    Wait until the message's timestamp + the deployments offset before handling
    """
    wait_minutes = int(env_vars()['TIMESTAMP_OFFSET_MINUTES'])
    if wait_minutes:
        message_process_time = (msg_timestamp // 1000) + (wait_minutes * 60)

        wait_time = message_process_time - datetime.datetime.timestamp(datetime.datetime.utcnow())

        if wait_time > 0:
            LOGGER.info(f'Waiting {wait_time} seconds before retry message processing continues; GUID {guid}')
            time.sleep(wait_time)


def poll_for_message(consumer):
    """
    Polls the broker for a message using the given timeout.
    If there are no messages to consume, either because None is returned
    or the message error is the no messages error,
    raises a NoMessageError.
    """
    message = consumer.poll(int(env_vars()['CONSUMER_POLL_TIMEOUT']))
    if message is None:
        raise NoMessageError
    return message


def handle_consumed_message(message, monitor):
    """
    Handles a consumed message to check for errors, handle retry waits, and log the consumption as a metric

    If the message is returned with a breaking error,
    raises a ConsumeMessageError.

    If the message is valid, waits until the message's timestamp plus
    the current process's time offset before handling the message.
    This allows retry deployments to wait in a non-
    blocking fashion
    """
    guid = None
    try:
        guid = [item[1] for item in message.headers() if item[0] == 'guid'][0].decode()
        LOGGER.info(f"Message consumed from topic {message.topic()}; GUID {guid}")
    except AttributeError:
        if "object has no attribute 'headers'" in str(message.error()):
            LOGGER.info("Message consumed. No headers, so no guid is available to log.")
    except TypeError:
        LOGGER.info("Message consumed. Headers are None, so no guid is available to log.")
    except IndexError:
        LOGGER.info("Message consumed. Headers found, but no guid is available to log.")

    # If message is None, it can mean that the poll operation timed out,
    # or that there are no more messages to consume

    if message.error():
        if "Broker: No more messages" in str(message.error()):
            LOGGER.warning("Consumer error: %s", message.error())
            raise NoMessageError(message.error())
        else:
            raise ConsumeMessageError(message.error())

    # Wait until message time if using a retry process
    _wait_until_message_time(message.timestamp()[1], guid)

    # Increment the metric for consumed messages by one
    monitor.inc_messages_consumed(1, message.topic())


def consume_message(consumer, monitor: MetricsManager):
    """
    Consumes a message from the broker while handling errors and waiting if necessary

    If the message is valid, then the message is returned
    """
    message = poll_for_message(consumer)
    handle_consumed_message(message, monitor)
    return message


def enqueue_auto_commits(consumer, messages):
    """
    A convenience method for staging message(s) to be committed automatically.
    """
    if env_vars()['CONSUMER_ENABLE_AUTO_COMMIT'].lower() == 'false':
        raise IncompatibleEnvironmentError('CONSUMER_ENABLE_AUTO_COMMIT', 'false')
    if isinstance(messages, list):
        for message in messages:
            consumer.store_offsets(message)
    else:
        consumer.store_offsets(messages)


def consume_message_batch(consumer, monitor: MetricsManager, count, timeout=30):
    """
    Consume a batch of messages.

    Requires manual deserialization since consumer.poll() is the method that has deserialization built in on it.
    """
    messages = consumer.consume(num_messages=count, timeout=timeout)
    if not messages:
        raise NoMessageError
    for message in messages:
        message.set_key(consumer._serializer.decode_message(message.key(), is_key=True))
        message.set_value(consumer._serializer.decode_message(message.value(), is_key=False))
        handle_consumed_message(message, monitor)
    return messages



import logging
from .producer_utils import confirm_produce
from .confluent_runtime_vars import env_vars
from nubium_utils.custom_exceptions import IncompatibleEnvironmentError

LOGGER = logging.getLogger(__name__)


def success_headers(headers):
    """
    Updates headers for a message when a process succeeds

    Resets the kafka_retry_count to 0,
    or adds if it didn't originally exist
    """

    try:
        headers = dict(headers)
    except TypeError:
        headers = {}

    headers['kafka_retry_count'] = '0'
    return headers


def produce_message_callback(error, message):
    """
    Logs the returned message from the Broker after producing
    NOTE: Headers not supported on the message for callback for some reason.
    NOTE: Callback requires these args
    """
    LOGGER.debug('Producer Callback...')
    if error:
        LOGGER.critical(error)
    else:
        LOGGER.debug('Producer Callback - Message produced successfully')


def consume_message_callback(error, partitions):
    """
    Logs the info returned when a successful commit is performed
    NOTE: Callback requires these args
    """
    LOGGER.debug('Consumer Callback...')
    if error:
        LOGGER.critical(error)
    else:
        LOGGER.debug('Consumer Callback - Message consumption committed successfully')


def synchronous_message_handling(producer=None, consumer=None):
    """
    Force the producer/consumer to acknowledge any outstanding, uncommitted messages.
    """
    if env_vars()['CONSUMER_ENABLE_AUTO_COMMIT'].lower() == 'true':
        raise IncompatibleEnvironmentError('CONSUMER_ENABLE_AUTO_COMMIT', 'true')
    if producer:
        LOGGER.debug('Waiting for produce to finish...')
        confirm_produce(producer)
    if consumer:
        LOGGER.debug('Waiting for commits to finalize...')
        consumer.commit(asynchronous=False)


def handle_no_messages(no_msg_exception=None, producer=None):
    """
    Since producer.poll() is typically called within the NU produce_message method, we want to acknowledge
    any outstanding produce attempts while the app has nothing to consume (and thus not actively producing/polling).
    """
    if no_msg_exception:
        LOGGER.debug(no_msg_exception)
    if producer:
        LOGGER.debug('flushing remaining producer queue')
        confirm_produce(producer)


def shutdown_cleanup(producer=None, consumer=None):
    """
    As part of shutdown, make sure all queued up produce messages are flushed, gracefully kill the consumer.
    """
    LOGGER.info('Performing graceful teardown of producer and/or consumer...')
    if consumer:
        LOGGER.debug("Shutting down consumer; no further commits can be queued or finalized.")
        consumer.close()
    if producer:
        LOGGER.debug("Sending/confirming the leftover messages in producer message queue")
        confirm_produce(producer, timeout=30)



class NoMessageError(Exception):
    """
    No message was returned when consuming from the broker
    """
    def __init__(self, *args):
        if not args:
            default_message = "No messages to consume"
            args = (default_message,)
        super().__init__(*args)


class NetworkError(Exception):
    """
    Generic handler for non-breaking, network-related external call failure.
    """
    def __init__(self, *args, req_status_code=None):
        self.req_status_code = 500
        if req_status_code:
            self.req_status_code = req_status_code
        if not args:
            default_message = f'Encountered a network-related external call failure, status code {self.req_status_code})'
            args = (default_message,)
        super().__init__(*args)


class ConsumeMessageError(Exception):
    """
    Non-trivial errors when consuming from the broker
    """
    pass


class MessageValueException(Exception):
    """
    Represents a significant error in the value of the message
    """
    pass


class ProduceHeadersException(Exception):
    """
    Represents a significant error in the value of the message
    """
    pass


class ProducerTimeoutFailure(Exception):
    """ Exception signifying a producer failing to communicate with the broker(s). """
    def __init__(self, *args):
        if not args:
            default_message = "Producer could not reach the broker(s) when attempting to send a message."
            args = (default_message,)
        super().__init__(*args)


class MaxRetriesReached(Exception):
    """
    Represents that all retry attempts for a retry-looped app have failed.
    """
    def __init__(self, *args):
        if not args:
            default_message = "All app retry attempts have failed"
            args = (default_message,)
        super().__init__(*args)


class RetryTopicSend(Exception):
    """
    Represents that an external call has failed but further retries will be attempted.
    Only a placeholder for logging a message sent to a retry topic when the app did not provide an explicit exception.
    """
    def __init__(self, *args):
        if not args:
            default_message = "External call attempt has failed, but will retry."
            args = (default_message,)
        super().__init__(*args)


class FailureTopicSend(Exception):
    """
    Represents that an external call has failed and no further retries will be attempted.
    Only a placeholder for logging a message sent to a failure topic when the app did not provide an explicit exception.
    """
    def __init__(self, *args):
        if not args:
            default_message = "External call attempt has failed and no (additional?) retries will be attempted."
            args = (default_message,)
        super().__init__(*args)


class IncompatibleEnvironmentError(Exception):
    def __init__(self, env_var_name, env_var_value, *args):
        default_message = f"The method called is not compatible with the env var {env_var_name}=={env_var_value}"
        args = (default_message, *args)
        super().__init__(*args)



import unittest
from unittest import mock
import socket

from prometheus_client import CollectorRegistry

from nubium_utils.metrics import MetricsManager, MetricsPusher


class TestMonitor(unittest.TestCase):
    def setUp(self):
        self.metrics_manager = MetricsManager(
            job="test-job",
            app="test-app",
            registry=CollectorRegistry(),
            metrics_pusher=mock.MagicMock(),
        )

    def test_registry_is_intialized_when_created_without_registry_argument(self):
        assert self.metrics_manager.registry.__class__ == CollectorRegistry

    def test_gauges_are_created_with_specific_labels(self):
        assert self.metrics_manager.messages_consumed._labelnames == (
            "app",
            "job",
            "topic",
        )
        assert self.metrics_manager.message_errors._labelnames == (
            "app",
            "job",
            "exception",
        )
        assert self.metrics_manager.messages_produced._labelnames == (
            "app",
            "job",
            "topic",
        )

    def test_utility_functions_increment_specific_label_values_on_each_gauge(self):
        self.metrics_manager.inc_messages_consumed(1, "test-topic")
        assert (
            self.metrics_manager.messages_consumed._metrics[
                ("test-app", "test-job", "test-topic")
            ]._value._value
            == 1
        )

        self.metrics_manager.inc_messages_produced(1, "test-topic")
        assert (
            self.metrics_manager.messages_produced._metrics[
                ("test-app", "test-job", "test-topic")
            ]._value._value
            == 1
        )

    def test_increment_message_errors_increases_corresponding_exception_label(self):
        self.metrics_manager.inc_message_errors(ValueError("test-value-error"))
        assert (
            self.metrics_manager.message_errors._metrics[
                ("test-app", "test-job", "ValueError")
            ]._value._value
            == 1
        )

    def test_registered_custom_metrics_are_incrementable(self):
        self.metrics_manager.register_custom_metric("totally_legit", "See! It even has a legit description!")
        self.metrics_manager.inc_custom_metric("totally_legit")
        self.metrics_manager.inc_custom_metric("totally_legit", 7)
        assert (
            self.metrics_manager.custom_gauges["totally_legit"]._metrics[
                ("test-app", "test-job", "totally_legit")
            ]._value._value
            == 8
        )

    def test_unregistered_custom_metrics_raise_error(self):
        with self.assertRaisesRegex(ValueError, "hammer_time.*not registered") as exc:
            self.metrics_manager.inc_custom_metric("hammer_time")


class TestMetricsPusher(unittest.TestCase):
    """
    Tests for the class that manages pushing metrics to a metrics cache
    """

    def test_initialization(self):
        """
        Class should initialize with correct arguments
        """
        metrics_pusher = MetricsPusher(
            "test-job", "test-service-name", "test-service-port", "test-pod-port"
        )
        assert metrics_pusher is not None

    def test_set_gateways(self):
        """
        Set gateways returns 2 IP addresses when called with mocked values
        """

        metrics_pusher = MetricsPusher(
            "test-job", "test-service-name", "test-service-port", "test-pod-port"
        )

        # Mock out the socket.getaddrinfo call return value
        addrinfo_mock = [
            (
                socket.AddressFamily.AF_INET,
                socket.SocketKind.SOCK_DGRAM,
                17,
                "",
                ("127.0.0.1", 8080),
            ),
            (
                socket.AddressFamily.AF_INET,
                socket.SocketKind.SOCK_DGRAM,
                6,
                "",
                ("127.0.0.1", 8080),
            ),
            (
                socket.AddressFamily.AF_INET,
                socket.SocketKind.SOCK_DGRAM,
                17,
                "",
                ("127.0.0.2", 8080),
            ),
            (
                socket.AddressFamily.AF_INET,
                socket.SocketKind.SOCK_DGRAM,
                6,
                "",
                ("127.0.0.2", 8080),
            ),
        ]

        with mock.patch("nubium_utils.metrics.metrics_pusher.socket") as socket_patch:
            socket_patch.getaddrinfo.return_value = addrinfo_mock
            metrics_pusher.set_metrics_pod_ips()
            assert metrics_pusher.metrics_pod_ips == {
                "127.0.0.1:test-pod-port",
                "127.0.0.2:test-pod-port",
            }

    def test_push_metrics(self):
        """
        Assert that the push metrics function is called for every gateway
        """

        metrics_pusher = MetricsPusher(
            "test-job", "test-service-name", "test-service-port", "test-pod-port"
        )

        metrics_pusher.metrics_pod_ips = {
            "test-ip-address-1:test-port",
            "test-ip-address-2:test-port",
        }

        with mock.patch(
            "nubium_utils.metrics.metrics_pusher.push_to_gateway"
        ) as push_to_gateway_patch:
            registry = CollectorRegistry()
            metrics_pusher.push_metrics(registry)
            assert (
                unittest.mock.call(
                    "test-ip-address-1:test-port",
                    job="test-job",
                    registry=registry,
                    timeout=15,
                )
                in push_to_gateway_patch.call_args_list
            )
            assert (
                unittest.mock.call(
                    "test-ip-address-2:test-port",
                    job="test-job",
                    registry=registry,
                    timeout=15,
                )
                in push_to_gateway_patch.call_args_list
            )



from unittest import TestCase
from nubium_utils.env_var_generator import env_vars_creator

class TestEnvVarGenerator(TestCase):
    def setUp(self):
        self.dict_out = {
            "test_env_field": "test_env_value"
        }
        self.env_vars = env_vars_creator(self.dict_func)

    def dict_func(self):
        return self.dict_out

    def test_it_works(self):
        self.assertDictEqual(self.env_vars(), self.dict_out)

    def test_returns_same_python_object_each_call(self):
        self.assertEqual(id(self.env_vars()), id(self.env_vars()))  # also ensures subsequent calls work!



import setuptools

with open("README.md", "r") as file_obj:
    long_description = file_obj.read()

install_requires = []
with open("requirements.txt") as file_obj:
    for req in file_obj:
        if req:
            install_requires.append(req)

faust_require = ['faust[rocksdb]==1.10.4', 'python-schema-registry-client[faust]==1.2.1']
tests_require = ["nose", "twine"] + faust_require

packages = setuptools.find_packages()

setuptools.setup(
    name="nubium-utils",
    version="0.23.0",
    author="Edward Brennan",
    author_email="ebrennan@redhat.com",
    description="Some Kafka utility functions and patterns for the nubium project",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.corp.redhat.com/mkt-ops-de/nubium-utils.git",
    packages=packages,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={"dev": tests_require, "faust": faust_require},
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)
